//
// maplexer.l
// lex -Pmsyy -I -Omaplexer.c maplexer.l
//
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
// #include "map.h" --> Discontinued in 5.x
#include "mapserver.h"
// #include <assert.h>

#ifdef USE_INFORMIXSPATIAL

#define ENDIAN_TEST(__a) \
{ union { short __sh; \
          char __ch[2]; \
        } __endian; \
  __endian.__sh = 0; \
  __endian.__ch[1] = '\001'; \
  __a = 1 - (__endian.__sh & 0x01); }

#define text unsigned char /* clon of oracle type */

#include <infxcli.h>
#include <ctype.h>

// #define MS_SUCCESS 0
// #define MS_FAILURE 1
#define BUFFER_LEN 1000 
#define ERRMSG_LEN 200
#define OBJECT_SIZE                100000
#define ARRAY_SIZE                 1024
#define QUERY_SIZE                 1
#define TEXT_SIZE                  256

/* The geometry type enumeration. */
typedef enum
 {
  geomEmpty             =  0,
  geomPoint             =  4,
  geomPointM            =  5,
  geomPointZ            =  6,
  geomPointZM           =  7,
  geomMultiPoint        =  8,
  geomMultiPointM       =  9,
  geomMultiPointZ       = 10,
  geomMultiPointZM      = 11,
  geomLineString        = 12,
  geomLineStringM       = 13,
  geomLineStringZ       = 14,
  geomLineStringZM      = 15,
  geomPolygon           = 16,
  geomPolygonM          = 17,
  geomPolygonZ          = 18,
  geomPolygonZM         = 19,
  geomMultiLineString   = 20,
  geomMultiLineStringM  = 21,
  geomMultiLineStringZ  = 22,
  geomMultiLineStringZM = 23,
  geomMultiPolygon      = 24,
  geomMultiPolygonM     = 25,
  geomMultiPolygonZ     = 26,
  geomMultiPolygonZM    = 27
 } GeomType;

/* The shape type enumeration. */
typedef enum
{
    shpNil              =  0,
    shpPoint            =  1,
    shpPolyline         =  3,
    shpPolygon          =  5,
    shpMultiPoint       =  8,
    shpPointZ           =  9,
    shpPolylineZ        = 10,
    shpPointZM          = 11,
    shpPolylineZM       = 13,
    shpPolygonZM        = 15,
    shpMultiPointZM     = 18,
    shpPolygonZ         = 19,
    shpMultiPointZ      = 20,
    shpPointM           = 21,
    shpPolylineM        = 23,
    shpPolygonM         = 25,
    shpMultiPointM      = 28
} ShpType;

/* The point coordinate structure. */
typedef struct
 {
  double  x;
  double  y;
 } Point;

typedef struct
 {
  GeomType  type;
  int       num_points;    /* Total # of points in the geometry */
  int       num_parts;     /* Total # of parts */
  int       num_subparts;  /* Total # of subparts */
  int      *offsets;       /* The parts offsets index */
  int      *suboffsets;    /* The subpart offset index */
  Point    *pt;            /* The geometry's points array */
  double   *z;             /* The geometry's z values array */
  double   *m;             /* The geometry's m values array */
  int       alloc_size;    /* Keeps track of size of alloc'd pt/z/m arrays */
  int       sub_allocsize; /*   "         "         "       suboffsets "   */
 } Geometry;

typedef text item_text[TEXT_SIZE];
typedef item_text item_text_array[ARRAY_SIZE];
typedef item_text item_text_array_query[QUERY_SIZE];  

typedef struct 
 {
  /* Informix handlers */
  SQLHDBC	hdbc;
  SQLHENV	henv;
 } msInformixSpatialHandler;

typedef struct 
 {
  /* informix handlers */
  msInformixSpatialHandler *informixhandlers;
  /* query data */
  SQLHSTMT	hstmt;
  /*
  ODBCStmt          *stmthp;
  ODBCDescribe      *dschp;
  ODBCType          *tdo;
  */
  /* fetch data */
  int              rows_fetched;
  int              row_num;
  int              row;
  item_text_array  *items; /* items buffer */
  SDWORD           *items_len;
  item_text_array_query  *items_query; /* items buffer */
  SDWORD           *items_query_len;
  SQLINTEGER FAR   fetched_shape_len;
  char             fetched_shape_buf[OBJECT_SIZE];
  // char             *fetched_shape_buf;
 } msInformixSpatialLayerInfo;

/* local prototypes */
static void msSplitLogin(char *connection, char *username, char *password, char *dblink);
static int msSplitData(char *data, char *geometry_column_name, char *table_name, char *unique, char *srid, int *function);
static void msODBCCloseConnection(void *layerinfo);
static msInformixSpatialHandler *msODBCSetHandlers(char *username, char *password, char *dblink);
static void msODBCSetLayerInfo(msInformixSpatialLayerInfo *layerinfo);
static void msODBCDisconnect(msInformixSpatialLayerInfo *layerinfo);
static void msODBCCloseHandlers(msInformixSpatialHandler *hand);
static void msODBCCloseLayerInfo(msInformixSpatialLayerInfo *layerinfo);
static void msODBCClearLayerInfo(msInformixSpatialLayerInfo *layerinfo);
// static int msODBCGetOrdinates(msInformixSpatialLayerInfo *layerinfo, etc. etc.
// static int msODBCConvertCircle(pointObj *pt);
static void osShapeBounds(shapeObj *shp);
// static void osPointCluster( etc etc);
// static void osPoint(etc etc);
// static void osClosePolygon(etc etc);
// static void osRectangle(etc etc);
// static void osCircle(etc etc);
// static int osGetOrdinates(etc etc);

void  get_integer (
    char  **buf_ptr,
    int     swap_bytes,
    int     num_ints,
    int    *int_vals
    )
{
    int     i;
    char   *to;
    char   *from;

    if (swap_bytes)
    {
        /* We need to byte swap */
        from = *buf_ptr;
        to = (char *)int_vals;
        for (i = 0; i < num_ints; i++)
        {
            to[0] = from[3];
            to[1] = from[2];
            to[2] = from[1];
            to[3] = from[0];
            to   += 4;
            from += 4;
        }

    }
    else
    {
        /* Simply copy the data */
        memcpy (int_vals, *buf_ptr, num_ints * sizeof (int));
    }

    /* Advance the pointer */
    *buf_ptr += num_ints * sizeof (int);

    return;
}

SQLINTEGER checkError (SQLRETURN rc,
                       SQLSMALLINT handleType,
		       SQLHANDLE handle,
		       SQLCHAR* errmsg)
 {
  SQLRETURN retcode = SQL_SUCCESS;
  SQLSMALLINT errNum = 1;
  SQLCHAR sqlState[6];
  SQLINTEGER nativeError;
  SQLCHAR errMsg[ERRMSG_LEN];
  SQLSMALLINT textLengthPtr;

  if((rc!=SQL_SUCCESS)&&(rc!=SQL_SUCCESS_WITH_INFO))
   {
    while(retcode!=SQL_NO_DATA)
     {
      retcode=SQLGetDiagRec(handleType, handle, errNum, sqlState, &nativeError, errMsg, ERRMSG_LEN, &textLengthPtr);

      if(retcode==SQL_INVALID_HANDLE)
       {
        fprintf (stderr, "checkError function was called with an invalid handle!!\n");
        return 1;
       }
      if((retcode==SQL_SUCCESS)||(retcode==SQL_SUCCESS_WITH_INFO))
        fprintf(stderr, "ERROR: %d:  %s : %s \n", nativeError, sqlState, errMsg);
        errNum++;
     }

    fprintf (stderr, "%s\n", errmsg);
    return 1;   /* all errors on this handle have been reported */
   }
  else
    return 0;	/* no errors to report */
 }

static void osShapeBounds(shapeObj *shp)
 {
  int i, f;

  fprintf(stderr, "osShapeBounds\n");
  if((shp->numlines>0)&&(shp->line[0].numpoints>0))
   {
    shp->bounds.minx=shp->line[0].point[0].x;
    shp->bounds.maxx=shp->line[0].point[0].x;
    shp->bounds.minx=shp->line[0].point[0].y;
    shp->bounds.maxx=shp->line[0].point[0].y;
   }
  for(i=0; i<shp->numlines; i++) 
   {
    for(f=0; f<shp->line[i].numpoints; f++)
     {
      if(shp->line[i].point[f].x<shp->bounds.minx)
        shp->bounds.minx=shp->line[i].point[f].x;
      if(shp->line[i].point[f].x>shp->bounds.maxx)
        shp->bounds.maxx=shp->line[i].point[f].x;
      if(shp->line[i].point[f].y>shp->bounds.miny)
        shp->bounds.miny=shp->line[i].point[f].y;
      if(shp->line[i].point[f].y>shp->bounds.maxy)
        shp->bounds.maxy=shp->line[i].point[f].y;
     }
   }
 }
/* 
   break layer->connection (username/password@dblink) 
   into username, password and dblink 
*/
static void msSplitLogin(char *connection, char *username, 
                         char *password, char *dblink)
 {
  char *src, *tgt;

  // if(layer->debug)
    msDebug("msSplitLogin called with: (%s)\n", connection);
  /* clearup */
  *username=*password=*dblink=0;
  /* bad 'connection' */
  if(connection==NULL) 
    return;
  /* ok, split connection */
  for(tgt=username, src=connection; *src; src++, tgt++ )
    if(*src=='/'||*src=='@') 
      break;
    else 
      *tgt=*src;
  *tgt=0;
  if(*src=='/') 
   {
    for(tgt=password, ++src; *src; src++, tgt++)
      if(*src=='@') 
        break;
      else 
        *tgt=*src;
    *tgt=0;
   }
  if(*src=='@')
    strcpy(dblink, ++src);
 }

static int msSplitData(char *data, char *geometry_column_name, char *table_name, char *unique, char *srid, int *function)
 {
  char *tok_from="from";
  char *tok_using="using";
  char *tok_unique="unique";
  char *tok_srid="srid";
  char tok_function[11]="";
  int  parenthesis, len;
  char *src=data, *tgt, *flk;

  // if(layer->debug)
    msDebug("msSplitData called\n");
  /* clearup */
  *geometry_column_name=*table_name=0;
  
  /* bad 'data' */
  if(data==NULL)
    return 0;

  /* parsing 'geometry_column_name' */
  for(; *src && isspace(*src); src++); /* skip blanks */
  for(tgt=geometry_column_name; *src; src++, tgt++)
    if(isspace(*src))
      break;
    else
      *tgt=*src;
  *tgt=0;

  /* parsing 'from' */
  for(; *src && isspace(*src); src++); /* skip blanks */
  for(; *src && *tok_from && tolower(*src)==*tok_from; src++, tok_from++);
  if(*tok_from!='\0')
    return 0;

  /* parsing 'table_name' or '(SELECT stmt)' */
  for(; *src && isspace(*src); src++); /* skip blanks */
  for(tgt=table_name, parenthesis=0; *src; src++, tgt++)
   {
    if(*src=='(')
      parenthesis++;
    else
      if(*src==')')
        parenthesis--;
      else
        if(parenthesis==0 && isspace(*src))
          break; /* stop on spaces */
    *tgt=*src;
   }
  *tgt=0;

  strcpy(unique, "");
  strcpy(srid, "NULL");
 
  /* parsing 'unique' */
  for(; *src && isspace(*src); src++); /* skip blanks */
  if(*src!='\0')
   {
    /* expects using [unique <column>] [srid <srid>] */
    /* parse 'using' */
    for(; *src && *tok_using && tolower(*src)==*tok_using; src++, tok_using++);
    if(*tok_using!='\0')
     {
      // msDebug("no encontre using\n");
      return 0;
     }
    /* parsing 'unique' */
    for(; *src && isspace(*src); src++); /* skip blanks */
    for(; *src && *tok_unique && tolower(*src)==*tok_unique; src++, tok_unique++);
    if(*tok_unique=='\0')
     {
      for(; *src && isspace(*src); src++); /* skip blanks */
      if(*src=='\0') 
        return 0;
      for(tgt=unique; *src; src++, tgt++)
        if(isspace(*src))
          break;
        else
          *tgt=*src;
      *tgt=0;
     } 
    /*
    else
      msDebug("no encontre unique\n");
    */
    /* parsing 'srid' */
    for(; *src && isspace(*src); src++); /* skip blanks */
    for(; *src && *tok_srid && tolower(*src)==*tok_srid; src++, tok_srid++);
    if(*tok_srid=='\0')
     {
      for(; *src && isspace(*src); src++); /* skip blanks */
      if(*src=='\0')
        return 0;
      for(tgt=srid; *src; src++, tgt++)
        if(isspace(*src))
          break;
        else
          *tgt=*src;
      tgt=0;
     }
    /* parsing function */
    for(; *src && isspace(*src); src++);
    flk=src;
    for(; *src && !isspace(*src); src++);
    len=src-flk;
    if(len>10)
      strcpy(tok_function, "ERROR");
    else
      strncpy(tok_function, flk, len);
   }
  // msDebug("geometry_column_name=(%s) table_name=(%s) unique=(%s)\n", geometry_column_name, table_name, unique);
  return(1); // ver que devuelvo
 }

static void msODBCCloseConnection(void *hand)
 {
  // if(layer->debug)
    msDebug("msODBCCloseConnection called\n");
  msODBCCloseHandlers((msInformixSpatialHandler *)hand);
 }

static msInformixSpatialHandler *msODBCSetHandlers(char *username, char *password, char *dblink)
 {
  SQLRETURN	rc=0;
  SQLCHAR	dsn[20];
  SQLCHAR	connStrIn[BUFFER_LEN];
  SQLCHAR	connStrOut[BUFFER_LEN];
  SQLSMALLINT	connStrOutLen;
  msInformixSpatialHandler *hand;
  msInformixSpatialLayerInfo *layerinfo;

  SQLCHAR datasource;
  SQLHSTMT hstmt;
  char    sql_stmt[200];
  int     rows_fetched;

  SQLSMALLINT   errNum=1;
  SQLCHAR       sqlState[6];
  SQLCHAR       errMsg[256];
  SQLSMALLINT   textLengthPtr;
  SQLINTEGER    nativeError;

  // if(layer->debug)
    msDebug("msODBCSetHandlers called (%s/%s@%s)\n", username, password, dblink);

  hand=(msInformixSpatialHandler *)malloc(sizeof(msInformixSpatialHandler));
  memset(hand, 0, sizeof(msInformixSpatialHandler)); 

  strcpy((char *)dsn, dblink);
  rc=SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &hand->henv);
  msDebug("SQLAllocHandle=%d\n", rc);
  rc=SQLSetEnvAttr(hand->henv, SQL_ATTR_ODBC_VERSION, (SQLPOINTER)SQL_OV_ODBC3, 0);
  msDebug("SQLSetEnvAttr=%d\n", rc);
  rc=SQLAllocHandle(SQL_HANDLE_DBC, hand->henv, &hand->hdbc);
  msDebug("SQLAllocHandle=%d\n", rc);
  rc=SQLConnect(hand->hdbc, (SQLCHAR *)dsn, SQL_NTS, (SQLCHAR *)"informix", SQL_NTS, (SQLCHAR *)"root", SQL_NTS);
  msDebug("SQLConnect=%d\n", rc);
  if(rc!=0)
   {
    SQLGetDiagRec(SQL_HANDLE_DBC, hand->hdbc, errNum, sqlState, &nativeError, errMsg, 256, &textLengthPtr);
    msDebug("ERROR: %d: %s %s\n", nativeError, sqlState, errMsg);
    exit(1);
   }
  return hand;
 }

static void msODBCSetLayerInfo(msInformixSpatialLayerInfo *layerinfo)
 {
  msInformixSpatialHandler *hand;
  SQLRETURN rc;
  char row_len[64];

  // if(layer->debug)
    msDebug("msODBCSetLayerInfo called\n");
  hand=(msInformixSpatialHandler *)layerinfo->informixhandlers;

  /* Allocate the statement handle */
  rc=SQLAllocHandle(SQL_HANDLE_STMT, hand->hdbc, &layerinfo->hstmt);
  // msDebug("SQLAllocHandle=%d\n", rc);
  // strcpy(row_len, "20000");
  // rc=SQLSetStmtAttr(&layerinfo->hstmt, SQL_ATTR_MAX_LENGTH, &row_len, sizeof(row_len));
  // msDebug("SQLSetStmtAttr=%d\n", rc);
 }

int msInformixSpatialLayerOpen(layerObj *layer)
 {
  char username[1024];
  char password[1024];
  char dblink[1024];

  msInformixSpatialLayerInfo *layerinfo=(msInformixSpatialLayerInfo *)malloc(sizeof(msInformixSpatialLayerInfo));
  // layerinfo->fetched_shape_buf=(char *)malloc(20000);
  msInformixSpatialHandler *hand=(msInformixSpatialHandler *)malloc(sizeof(msInformixSpatialHandler));
  memset(hand, 0, sizeof(msInformixSpatialHandler));
  memset(layerinfo, 0, sizeof(msInformixSpatialLayerInfo));
  // if(layer->debug)
    msDebug("InformixSpatialLayerOpen called with: (%s)\n", layer->data);
  if(layer->data==NULL)
   {
    msSetError(MS_INFORMIXSPATIALERR, 
      "Error parsing InformixSpatial DATA variable. Must be:"
      "'geometry_column FROM table_name [USING UNIQUE <column> SRID srid# FUNCTION]'.",
      "msInformixSpatialLayerOpen()"); 
    return MS_FAILURE;
   }
  msSplitLogin(layer->connection, username, password, dblink);
  layerinfo->informixhandlers=(msInformixSpatialHandler *)msConnPoolRequest(layer);
  if(layerinfo->informixhandlers==NULL)
   {
    layerinfo->informixhandlers=msODBCSetHandlers(username, password, dblink);
    if(layerinfo->informixhandlers==NULL)
     {
      msSetError(MS_INFORMIXSPATIAL, 
        "Cannot create ODBC Handlers. "
        "Connection failure. Check the connection string. "
        "Error: %s. ",
        "msInformixSpatialLayerOpen()", 0);
      msODBCClearLayerInfo(layerinfo);
      msODBCCloseLayerInfo(layerinfo);
      return MS_FAILURE;
     }
    msConnPoolRegister(layer, layerinfo->informixhandlers, msODBCCloseConnection);
   }
  msODBCSetLayerInfo(layerinfo);
  if(layerinfo==NULL)
   {
    msSetError(MS_INFORMIXSPATIALERR, 
      "Cannot create ODBC LayerInfo. "
      "Connection failure.",
      "msInformixSpatialLayerOpen()");
    msODBCClearLayerInfo(layerinfo);
    msODBCCloseHandlers(layerinfo->informixhandlers);
    msODBCCloseLayerInfo(layerinfo);
    return MS_FAILURE;
   }
  layer->layerinfo=layerinfo;
  return layer->layerinfo!=NULL ? MS_SUCCESS : MS_FAILURE;
 }

int msInformixSpatialLayerIsOpen(layerObj *layer)
 {
  // if(layer->debug)
    msDebug("msInformixSpatialLayerIsOpen called\n");
  return MS_SUCCESS;
 }

int msInformixSpatialLayerClose(layerObj *layer)
 {
  msInformixSpatialLayerInfo *layerinfo;

  // if(layer->debug)
    msDebug("msInformixSpatialLayerClose was called. Layer connection: %s\n", layer->connection);
  layerinfo=((msInformixSpatialLayerInfo *)layer->layerinfo);
  if(layerinfo!=NULL)
   {
    // if(layer->debug)
      msDebug("msInformixSpatialLayerClose. Cleaning layerinfo handlers.\n");
    msODBCClearLayerInfo(layerinfo);
    // if(layer->debug)
      msDebug("msInformixSpatialLayerClose. Cleaning Informix handlers.\n");
    msConnPoolRelease(layer, layerinfo->informixhandlers);
    layerinfo->informixhandlers=NULL;
    msODBCCloseLayerInfo(layerinfo);
    layer->layerinfo=NULL;
   }
  return MS_SUCCESS;
 }

int msInformixSpatialLayerWhichShapes(layerObj *layer, rectObj rect)
 {
  SQLRETURN	rc=0;
  msInformixSpatialLayerInfo *layerinfo;
  msInformixSpatialHandler *hand;
  int success, i;
  int function=0;
  char table_name[2000], geom_column_name[100], unique[100], srid[100];
  char query_str[6000];

  char    sql_stmt[200];
  int     rows_fetched;

  /* get layerinfo */
  layerinfo=(msInformixSpatialLayerInfo *)layer->layerinfo;
  hand=(msInformixSpatialHandler *)layerinfo->informixhandlers;
  // if(layer->debug)
    msDebug("msInformixSpatialLayerWhichShapes called.\n");
  if(layerinfo==NULL)
   {
    msSetError(MS_INFORMIXSPATIALERR, 
      "msInformixSpatialLayerWhichShapes calles on unopened layer",
      "msInformixSpatialLayerWhichShapes()");
    return MS_FAILURE;
   }
  /* parse goem_column_name and table_name */
  if(!msSplitData(layer->data, geom_column_name, table_name, unique, srid, &function))
   {
    msSetError(MS_INFORMIXSPATIALERR,
      "Error parsing InformixSpatial DATA variable. Must be:"
      "'geometry_column FROM table_name [USING UNIQUE <column> SRID srid# FUNCTIOn]' or "
      "'geometry_column FROM (SELECT stmt) [USING UNIQUE <column> SRID srid# FUNCTION]'."
      "If want to set the FUNCTION statement you can use: FILTER, RELATE, GEOMRELATE or NONE."
      "Your data statement: %s",
      "msInformixSpatialLayerWhichSpapes()", layer->data);
    return MS_FAILURE;
   }
  /* Define the unique */
  if(unique[0]=='\0')
    strcpy(unique, "se_row_id");
  sprintf(query_str, "SELECT %s", unique);
  /* allocate enough space for items */
  if(layer->numitems>=0)
   {
    layerinfo->items=(item_text_array *)malloc(sizeof(item_text_array)*(layer->numitems+1));
    if(layerinfo->items==NULL)
     {
      msSetError(MS_INFORMIXSPATIALERR, 
        "Cannot allocate items buffer",
        "msInformixSpatialLayerWhichShapes()");
      return MS_FAILURE;
     }
    layerinfo->items_len=(SDWORD *)malloc(sizeof(SDWORD)*(layer->numitems+1));
   }
  /* define SQL query */
  // msDebug("numitems=%d\n", layer->numitems);
  for(i=0; i<layer->numitems; ++i)
    sprintf(query_str+strlen(query_str), ", %s", layer->items[i]);
  sprintf(query_str+strlen(query_str), ", SE_AsSHAPE(%s) FROM %s", geom_column_name, table_name);
  /*
  if(layer->filter.string!=NULL)
   {
   }
  */
  // msDebug("query_str=%s\n", query_str);
  rc=SQLPrepare(layerinfo->hstmt, (UCHAR *)query_str, SQL_NTS);
  // msDebug("SQLPrepare=%d\n", rc);
  rc=SQLExecute(layerinfo->hstmt);
  // msDebug("SQLExecute=%d\n", rc);
  for(i=0; i<=layer->numitems; i++)
   {
    layerinfo->items_len[i]=0;
    rc=SQLBindCol(layerinfo->hstmt, i+1, SQL_C_CHAR, (SQLPOINTER)layerinfo->items[i], (SQLINTEGER)(TEXT_SIZE), &layerinfo->items_len[i]);
    // msDebug("i=%d SQLBindCol=%d\n", i, rc);
   }
  layerinfo->fetched_shape_len=0;
  rc=SQLBindCol(layerinfo->hstmt, layer->numitems+2, SQL_C_BINARY, (SQLPOINTER)layerinfo->fetched_shape_buf, (SQLINTEGER)(OBJECT_SIZE), &layerinfo->fetched_shape_len);
  // msDebug("i=%d SQLBindCol=%d\n", i, rc);
  layerinfo->row_num=layerinfo->row=0;

  return MS_SUCCESS;
 }

/* get ordinates from buffer */
static int msODBCGetOrdinates(msInformixSpatialLayerInfo *layerinfo, int s, pointObj *pt, int num_points)
 {
  double x, y;
  int i, n, success=1;
  // boolean exists;
  Point   *points;
  char    *cp;
  int     swap_bytes;
  msInformixSpatialHandler *hand;
  int     et;
  
  // if(layer->debug)
    msDebug("msODBCGetOrdinates called\n");
  hand=(msInformixSpatialHandler *)layerinfo->informixhandlers;
  points=(Point *)malloc(sizeof(Point));

  for(i=s, n=0; n<num_points; i=i+sizeof(double)*2, n++)
  // for(i=s, n=num_points-1; n>=0; i=i+sizeof(double)*2, n--)
   {
    // cp=&layerinfo->fetched_shape_buf[0];
    // ENDIAN_TEST(i);
    // swap_bytes=1-i;
 
    bzero(points, sizeof(Point)); 
    memcpy(points, &layerinfo->fetched_shape_buf[i], 1*sizeof(Point));
    // msDebug("%d: num_points=%d Points=%.9f %.9f\n", n, num_points, points->x, points->y);
    pt[n].x=points->x;
    pt[n].y=points->y;
   }
  // msDebug("salgo de msODBCGetOrdinates con n=%d\n", n);
  free(points);
  return(n);
  // return(num_points);
 }

static void osPoint(msInformixSpatialLayerInfo *layerinfo, shapeObj *shape, int start, lineObj points, pointObj *pnt)
 {
  int n;
  // 
  // ojo con el "1" !!!
  // 
    msDebug("msInformixSpatialLayerInfo called\n");
  n=msODBCGetOrdinates(layerinfo, start, pnt, 1);
  if(n==1)
   {
    shape->type=MS_SHAPE_POINT;
    points.numpoints=1;
    points.point=pnt;
    msAddLine(shape, &points);
   }
 }

static void osClosedPolygon(msInformixSpatialLayerInfo *layerinfo, shapeObj *shape, int start, lineObj points, int elem_type, int num_points)
 {
  int n;

  // if(layer->debug)
    msDebug("osClosedPolygon called (%d points)\n", num_points);
  points.point=(pointObj *)malloc(sizeof(pointObj)*num_points);
  n=msODBCGetOrdinates(layerinfo, start, points.point, num_points);
  // msDebug("num_points=%d n=%d\n", num_points, n);
  if(n>0)
   {
    //
    // ver aqui!!!
    //
    shape->type=(elem_type==3)?MS_SHAPE_LINE:MS_SHAPE_POLYGON;
    points.numpoints=n;
    msAddLine(shape, &points);
   }
  free(points.point);
 }

static int osGetOrdinates(msInformixSpatialLayerInfo *layerinfo, shapeObj *shape)
 {
  ShpType shape_type;
  int     elem_type;
  int     elem, ord_start;
  lineObj points={0, NULL};
  pointObj point5[5]; 

  char    *cp;
  int     swap_bytes;
  int     num_parts, num_points, np;
  int     *offsets;
  int     i;

  Point   *wpoints;

  // if(layer->debug)
    msDebug("osGetOrdinates called\n");

  elem_type=layerinfo->fetched_shape_buf[0];
  // msDebug("Type=%d\n", elem_type);
  if(elem_type==1)
   {
    wpoints=(Point *)malloc(sizeof(Point));
    memcpy(wpoints, &layerinfo->fetched_shape_buf[4], 1*sizeof(Point));
    // msDebug("Points=%.0f %.0f\n", wpoints->x, wpoints->y);

    shape->type=MS_SHAPE_POINT;
    points.numpoints=1;
    points.point=point5;
    point5[0].x=wpoints->x;
    point5[0].y=wpoints->y;
    msAddLine(shape, &points);
    free(wpoints);
    return MS_SUCCESS;
   }

  switch(elem_type)
   {
    case 3: /* linestring type */
      cp=&layerinfo->fetched_shape_buf[0];
      ENDIAN_TEST(i);
      swap_bytes=1-i;
      get_integer(&cp, swap_bytes, 1, (int *)&elem_type);
      cp+=4*sizeof(double);
      get_integer(&cp, swap_bytes, 1, &num_parts);
      get_integer(&cp, swap_bytes, 1, &num_points);

      // msDebug(">>>num_points=%d\n", num_points);

      offsets=malloc(num_parts*sizeof(int));
      get_integer(&cp, swap_bytes, num_parts, offsets);
      for(i=0; i<num_parts; i++)
       {
        // msDebug("offset %d=%d\n", i, offsets[i]);
        ord_start=(int)(cp-layerinfo->fetched_shape_buf+offsets[i]*sizeof(double)*2);
        if(i==(num_parts-1))
          np=num_points;
        else
         {
          // num_points-=offsets[i+1];
          np=offsets[i+1]-offsets[i];
          num_points-=np;
         }
        // ord_end=ord_start+num_points*sizeof(double);
        // msDebug("num_parts=%d num_points=%d np=%d\n", num_parts, num_points, np);
        // msDebug("ord_start=%d\n", ord_start);
        osClosedPolygon(layerinfo, shape, ord_start, points, elem_type, np);
       }
      free(offsets);
      break;
    case 5: /* multipolygon type */
      cp=&layerinfo->fetched_shape_buf[0];
      ENDIAN_TEST(i);
      swap_bytes=1-i;
      get_integer(&cp, swap_bytes, 1, (int *)&elem_type);
      cp+=4*sizeof(double);
      get_integer(&cp, swap_bytes, 1, &num_parts);
      get_integer(&cp, swap_bytes, 1, &num_points);
      offsets=malloc(num_parts*sizeof(int));
      get_integer(&cp, swap_bytes, num_parts, offsets);
      for(i=0; i<num_parts; i++)
       {
        // msDebug("offset %d=%d\n", i, offsets[i]);
        ord_start=(int)(cp-layerinfo->fetched_shape_buf+offsets[i]*sizeof(double)*2);
        if(i==(num_parts-1))
          np=num_points;
        else
         {
          num_points-=offsets[i+1];
          np=offsets[i+1]-offsets[i];
         }
        // ord_end=ord_start+num_points*sizeof(double);
        // msDebug("num_parts=%d num_points=%d np=%d\n", num_parts, num_points, np);
        // msDebug("ord_start=%d\n", ord_start);
        osClosedPolygon(layerinfo, shape, ord_start, points, elem_type, np);
       }
      free(offsets);
      break;
    case 4: /* point type */
      osPoint(layerinfo, shape, ord_start, points, point5);
      break;
   }
  return MS_SUCCESS;
 }

int msInformixSpatialLayerNextShape(layerObj *layer, shapeObj *shape)
 {
  msInformixSpatialLayerInfo *layerinfo;
  msInformixSpatialHandler *hand;
  SQLRETURN	rc=0;
  int     i;
  int     success;
  Geometry *obj;

  SQLCHAR     wcolumn[300];
  SQLSMALLINT wcolumn_len;
  SQLSMALLINT wcolumn_size;
  SQLSMALLINT wcolumn_type;
  SQLUINTEGER wcolumn_prec;
  SQLSMALLINT wcolumn_scale;
  SQLSMALLINT wcolumn_null;

  unsigned char wstate[6];
  SDWORD        werr;
  unsigned char werr_msg[SQL_MAX_MESSAGE_LENGTH];
  SWORD         wlength;

  // if(layer->debug)
    msDebug("msInformixSpatialLayerNextShape called\n");
  layerinfo=(msInformixSpatialLayerInfo *)layer->layerinfo;
  hand=(msInformixSpatialHandler *)layerinfo->informixhandlers;
  if(layerinfo==NULL)
   {
    msSetError(MS_INFORMIXSPATIALERR, "msInformixSpatialLayerNextShapes called on unopened layer", "msInformixSpatialLayerNextShapes()");
    return MS_FAILURE;
   }
  for(;;)
   {
    // msDebug("Antes de SQLFetch\n");
    rc=SQLFetch(layerinfo->hstmt);
    // msDebug("SQLFetch=%d(%d)\n", rc, SQL_SUCCESS);
    // msDebug("fetched_shape_len=%d record=%ld\n", layerinfo->fetched_shape_len, atol(layerinfo->items[0][0]));
    if(rc!=0)
     {
      SQLGetDiagRec(SQL_HANDLE_STMT, layerinfo->hstmt, 1, wstate, &werr, &werr_msg, SQL_MAX_MESSAGE_LENGTH-1, &wlength);
      // msDebug("Error=(%d)(%s)\n", werr, werr_msg);
      // return MS_FAILURE;
     }
    if(layerinfo->fetched_shape_len==0)
     {
      // msDebug("MS_DONE\n");
      return MS_DONE;
     }
    if((rc!=SQL_SUCCESS)&&(werr!=-11003))
     {
      // msDebug("MS_DONE\n");
      return MS_DONE;
     }
    if(rc==0)
     {
      // msDebug("rc=0");
      break;
     }
    // msDebug("loop\n");
   }
  layerinfo->row=0;

  // obj=layerinfo->obj[layerinfo->row];
  // ind=layerinfo->ind[layerinfo->row];

  /* get the items for the shape */
  shape->index=atol(layerinfo->items[0][layerinfo->row]);
  shape->numvalues=layer->numitems;
  // msDebug("index=%ld numvalues=%ld\n", shape->index, layer->numitems);
  shape->values=(char **)malloc(sizeof(char *)*shape->numvalues);
  for(i=0; i<shape->numvalues; ++i)
    shape->values[i]=strdup(layerinfo->items[i+1][layerinfo->row]);

  /* increment for next row */
  layerinfo->row_num++;
  layerinfo->row++;

  /* fetch a layer->type object */
  success=osGetOrdinates(layerinfo, shape);
  // msDebug("success=%d MS_SUCCESS=%d\n", success, MS_SUCCESS); 
  if(success!=MS_SUCCESS)
    return MS_FAILURE;
  osShapeBounds(shape);
  return MS_SUCCESS;
 }

int msInformixSpatialLayerGetItems(layerObj *layer)
 {
  char *rzt="";
  // char *flk="";
  char flk[2048];
  int  function=0;
  int  existgeom;
  int  count_item, flk_len, success, i, j;
  char query_str[6000], table_name[2000], geom_column_name[100], unique[100], srid[100];
  msInformixSpatialLayerInfo *layerinfo;
  msInformixSpatialHandler *hand;
  SQLRETURN	rc=0;
  
  layerinfo=(msInformixSpatialLayerInfo *)layer->layerinfo;
  hand=(msInformixSpatialHandler *)layerinfo->informixhandlers;
  // if(layer->debug)
    msDebug("msInformixSpatialLayerGetItems was called.\n");

  if(layerinfo==NULL)
   {
    msSetError(MS_INFORMIXSPATIALERR, "msInformixSpatialLayerGetItems called on unopened layer", "msInformixSpatialLayerGetItems()");
    return MS_FAILURE;
   }
  if(!msSplitData(layer->data, geom_column_name, table_name, unique, srid, &function))
   {
    msSetError(MS_INFORMIXSPATIALERR, 
      "Error parsing InformixSpatial DATA variable. Must be:"
      "'geometry_column FROM table_name [USING UNIQUE <column> SRID srid# FUNCTION]'.",
      "msInformixSpatialLayerOpen()"); 
    return MS_FAILURE;
   } 
  sprintf(query_str, "SELECT * FROM %s", table_name);
  rc=SQLPrepare(layerinfo->hstmt, (UCHAR *)query_str, SQL_NTS);
  // msDebug("SQLPrepare de (%s)=%d\n", query_str, rc);
  // ver ver ver ver
  // rc=SQLExecute(layerinfo->hstmt);
  // msDebug("SQLExecute=%d\n", rc);
  rc=SQLNumResultCols(layerinfo->hstmt, &layer->numitems);
  // msDebug("SQLNumResultCols=%d numitems=%d\n", rc, layer->numitems);
  layer->numitems=layer->numitems-1;
  layer->items=malloc(sizeof(char *)*(layer->numitems));
  if(layer->numitems>0)
   {
    layerinfo->items_query=(item_text_array_query *)malloc(sizeof(item_text_array_query)*(layer->numitems));
    if(layerinfo->items_query==NULL)
     {
      msSetError(MS_INFORMIXSPATIALERR, "Cannot allocate items buffer", "msInformixSpatialLayerGetItems()");
      return MS_FAILURE;
     }
    layerinfo->items_query_len=(SDWORD *)malloc(sizeof(SDWORD)*(layer->numitems+1));
   }
  count_item=0;
  existgeom=0;
  /* Upcase conversion for the geom_column_name */
  for(i=0; geom_column_name[i]!=0; i++)
    geom_column_name[i]=toupper(geom_column_name[i]);
  // msDebug("geom_column_name=%s", geom_column_name);

  /* Retrieve columns name from the user table */
  for(i=0; i<=layer->numitems; i++)
   {
    flk_len=0;
    rc=SQLColAttribute(layerinfo->hstmt, i+1, SQL_DESC_BASE_COLUMN_NAME, &flk, sizeof(flk), &flk_len, NULL);
    flk[flk_len]=0;
    for(j=0; flk[j]!=0; j++)
      flk[j]=toupper(flk[j]);
    if(strcmp(flk, geom_column_name))
     {
      layer->items[count_item]=malloc(flk_len);
      strcpy(layer->items[count_item], flk);
      count_item++;
     } 
    else
      existgeom=1;
   }
  if(!existgeom)
   {
    msSetError(MS_INFORMIXSPATIALERR, "No geometry column, check stmt", "msInformixSpatialLayerGetItems()");
    return MS_FAILURE;
   }
  return msInformixSpatialLayerInitItemInfo(layer);
 }

int msInformixSpatialLayerGetShape(layerObj *layer, shapeObj *shape, long record)
 {
  char query_str[6000], table_name[2000], geom_column_name[100], unique[100], srid[100];
  int success, i;
  int function=0;
  SQLRETURN rc;
  msInformixSpatialLayerInfo *layerinfo;
  msInformixSpatialHandler *hand;
 
  SQLSMALLINT errNum = 1;
  SQLCHAR sqlState[6];
  SQLINTEGER nativeError;
  SQLCHAR errMsg[ERRMSG_LEN];
  SQLSMALLINT textLengthPtr;

  layerinfo=(msInformixSpatialLayerInfo *)layer->layerinfo;
  hand=(msInformixSpatialHandler *)layerinfo->informixhandlers;
  // if(layer->debug)
    msDebug("msInformixSpatialLayerGetShape called. Using the record = %ld.\n", record);
  if(layerinfo==NULL)
   {
    msSetError(MS_INFORMIXSPATIALERR, "msInformixSpatialLayerGetShape called on unopened layer", "msInformixSpatialLayerGetShape()");
    return MS_FAILURE;
   }
  /* allocate enough space for items */
  if(layer->numitems>0)
   {
    layerinfo->items_query=(item_text_array_query *)malloc(sizeof(item_text_array_query)*(layer->numitems));
    if(layerinfo->items_query==NULL)
     {
      msSetError(MS_INFORMIXSPATIALERR, "Cannot allocate items buffer", "msInformixSpatialLayerWhichShapes()");
      return MS_FAILURE;
     }
    layerinfo->items_query_len=(SDWORD *)malloc(sizeof(SDWORD)*(layer->numitems+1));
   }
  if(!msSplitData(layer->data, geom_column_name, table_name, unique, srid, &function))
   {
    msSetError(MS_INFORMIXSPATIALERR, 
      "Error parsing InformixSpatial DATA variable. Must be: "
      "'geometry_column FROM table_name [USING UNIQUE <column> SRID srid# FUNCTION]' or "
      "completar completar", "");
    return MS_FAILURE;
   }
  /* Define the first query to retrieve items */
  if(unique[0]==0)
   {
    msSetError(MS_INFORMIXSPATIALERR, "Error parsing InformixSpatial DATA variable for query. To execute "
      "query functions you need to define one "
      "unique column [USING UNIQUE <column>#]",
      "msInformixSpatialLayerGetShape()");
    return MS_FAILURE;
   }
  else
    sprintf(query_str, "SELECT ");
  /* Define the query */
  for(i=0; i<layer->numitems; i++)
    sprintf(query_str+strlen(query_str), " %s,", layer->items[i]);
  sprintf(query_str+strlen(query_str), " SE_AsSHAPE(%s) FROM %s WHERE %s = %ld", geom_column_name, table_name, unique, record);
  // if(layer->filter.string!=NULL)
  //   sprintf(query_str+strlen(query_str), " AND %s", (layer->filter.string));
  // if(layer->debug)
    msDebug("msInformixSpatialLayerGetShape. Sql: %s (%d)\n", query_str, layer->numitems);
  /* Prepare the handlers to the query */
  rc=SQLPrepare(layerinfo->hstmt, (UCHAR *)query_str, SQL_NTS);
  if(rc!=0)
   {
    SQLGetDiagRec(SQL_HANDLE_STMT, layerinfo->hstmt, errNum, sqlState, &nativeError, &errMsg, 256, &textLengthPtr);
    // msDebug("ERROR: %d: %s %s\n", nativeError, sqlState, errMsg);
   }
  // msDebug("SQLPrepare=%d\n", rc);
  rc=SQLExecute(layerinfo->hstmt);
  // msDebug("SQLExecute=%d\n", rc);
  for(i=0; i<layer->numitems; i++)
   {
    layerinfo->items_query_len[i]=0;
    rc=SQLBindCol(layerinfo->hstmt, i+1, SQL_C_CHAR, (SQLPOINTER)layerinfo->items_query[i], (SQLINTEGER)(TEXT_SIZE), &layerinfo->items_query_len[i]);
   }
  layerinfo->fetched_shape_len=0;
  rc=SQLBindCol(layerinfo->hstmt, layer->numitems+1, SQL_C_BINARY, (SQLPOINTER)layerinfo->fetched_shape_buf, (SQLINTEGER)(OBJECT_SIZE), &layerinfo->fetched_shape_len);
  rc=SQLFetch(layerinfo->hstmt);

  shape->type=MS_SHAPE_NULL;
  // if(layerinfo->rows_fetched==0)
  //   return(MS_DONE);
  shape->numvalues=layer->numitems;
  shape->values=(char **)malloc(sizeof(char *)*shape->numvalues);
  shape->index=record;
  for(i=0; i<shape->numvalues; ++i)
    shape->values[i]=strdup(layerinfo->items_query[layerinfo->row][i]);

  /* increment for the next row */
  layerinfo->row_num++;
  layerinfo->row++;

  /* fetch a layer->type object */
  success=osGetOrdinates(layerinfo, shape);
  // msDebug("success=%d MS_SUCCESS=%d\n", success, MS_SUCCESS); 
  if(success!=MS_SUCCESS)
    return MS_FAILURE;
  osShapeBounds(shape);
  layerinfo->row=layerinfo->row_num=0;
  return MS_SUCCESS;
 }

int msInformixSpatialLayerGetExtent(layerObj *layer, rectObj *extent)
 {
  msSetError(MS_INFORMIXSPATIALERR, "Function not implemented yet", "msInformixSpatialLayerGetExtent()"); 
  return MS_FAILURE;
 }

static void msODBCDisconnect(msInformixSpatialLayerInfo *layerinfo)
 {
  msInformixSpatialHandler *hand;

  // if(layer->debug)
    msDebug("msODBCDisconnect called\n");
  if(layerinfo->hstmt!=NULL)
   {
    SQLFreeStmt(layerinfo->hstmt, SQL_CLOSE);          // Close the statement handle
    SQLFreeHandle(SQL_HANDLE_STMT, layerinfo->hstmt);  // Free the statement handle 
   }
  hand=(msInformixSpatialHandler *)layerinfo->informixhandlers;
  if(hand->hdbc!=NULL)
   {
    SQLDisconnect(hand->hdbc);
    SQLFreeHandle(SQL_HANDLE_DBC, hand->hdbc);
    SQLFreeHandle(SQL_HANDLE_DBC, hand->henv);
   }
  memset(hand, 0, sizeof(msInformixSpatialHandler));
  if(layerinfo->items!=NULL)
    free(layerinfo->items);
  if(layerinfo->items_query!=NULL)
    free(layerinfo->items_query);
  memset(hand, 0, sizeof(msInformixSpatialHandler));
  memset(layerinfo, 0, sizeof(msInformixSpatialLayerInfo));
 }

static void msODBCCloseHandlers(msInformixSpatialHandler *hand)
 {
  // if(layer->debug)
    msDebug("msODBCCloseHandlers called\n");
  if(hand->hdbc!=NULL)
   {
    SQLDisconnect(hand->hdbc);
    SQLFreeHandle(SQL_HANDLE_DBC, hand->hdbc);
   }
  if(hand->henv!=NULL)
    SQLFreeHandle(SQL_HANDLE_DBC, hand->henv);
  if(hand!=NULL)
    memset(hand, 0, sizeof(msInformixSpatialHandler));
  free(hand);
 }

static void msODBCCloseLayerInfo(msInformixSpatialLayerInfo *layerinfo)
 {
  // if(layer->debug)
    msDebug("msODBCCloseLayerInfo called\n");
  if(layerinfo!=NULL)
    memset(layerinfo, 0, sizeof(msInformixSpatialLayerInfo));
  free(layerinfo);
 }

static void msODBCClearLayerInfo(msInformixSpatialLayerInfo *layerinfo)
 {
  // if(layer->debug)
    msDebug("msODBCClearLayerInfo called\n");

  if(layerinfo->hstmt!=NULL)
   {
    SQLFreeStmt(layerinfo->hstmt, SQL_CLOSE);          // Close the statement handle
    SQLFreeHandle(SQL_HANDLE_STMT, layerinfo->hstmt);  // Free the statement handle 
   } 
  if(layerinfo->items!=NULL)
    free(layerinfo->items);
  if(layerinfo->items_query!=NULL)
    free(layerinfo->items_query);
 }

int msInformixSpatialLayerInitItemInfo(layerObj *layer)
 {
  int i;
  int *itemindexes;

  // if(layer->debug)
    msDebug("msInformixSpatialLayerInitItemInfo called\n"); 
  msDebug("numitems=%d\n", layer->numitems); 
  
  if(layer->numitems==0)
    return MS_SUCCESS;
  if(layer->iteminfo)
    free(layer->iteminfo);
  if((layer->iteminfo=(long *)malloc(sizeof(int)*layer->numitems))==NULL)
   {
    msSetError(MS_MEMERR, NULL, "msInformixSpatialLayerInitItemInfo()");
    return MS_FAILURE;
   }
  itemindexes=(int *)layer->iteminfo;
  for(i=0; i<layer->numitems; i++)
    itemindexes[i]=i;
  return MS_SUCCESS;
 }

void msInformixSpatialLayerFreeItemInfo(layerObj *layer)
 {
  // if(layer->debug)
    msDebug("msInformixSpatialLayerFreeItemInfo was called.\n");
  if(layer->iteminfo)
    free(layer->iteminfo);
  layer->iteminfo=NULL;
  /* nothing to do */
 } 

#else

/* InformixSpatial "not-supported" procedures */

int msInformixSpatialLayerOpen(layerObj *layer)
 {
  msSetError(MS_INFORMIXSPATIALERR, "InformixSpatial is not supported", "msInformixSpatialLayerOpen()"); 
  return MS_FAILURE;
 }

int msInformixSpatialLayerIsOpen(layerObj *layer)
 {
  msSetError(MS_INFORMIXSPATIALERR, "InformixSpatial is not supported", "msInformixSpatialLayerIsOpen()"); 
  return MS_FAILURE;
 }

int msInformixSpatialLayerClose(layerObj *layer)
 {
  msSetError(MS_INFORMIXSPATIALERR, "InformixSpatial is not supported", "msInformixSpatialLayerClose()"); 
  return MS_FAILURE;
 }

int msInformixSpatialLayerWhichShapes(layerObj *layer, rectObj rect)
 {
  msSetError(MS_INFORMIXSPATIALERR, "InformixSpatial is not supported", "msInformixSpatialLayerWhichShapes()"); 
  return MS_FAILURE;
 }

int msInformixSpatialLayerNextShape(layerObj *layer, shapeObj *shape)
 {
  msSetError(MS_INFORMIXSPATIALERR, "InformixSpatial is not supported", "msInformixSpatialLayerNextShape()"); 
  return MS_FAILURE;
 }

int msInformixSpatialLayerGetItems(layerObj *layer)
 {
  msSetError(MS_INFORMIXSPATIALERR, "InformixSpatial is not supported", "msInformixSpatialLayerGetItems()"); 
  return MS_FAILURE;
 }

int msInformixSpatialLayerGetShape(layerObj *layer, shapeObj *shape, long record)
 {
  msSetError(MS_INFORMIXSPATIALERR, "InformixSpatial is not supported", "msInformixSpatialLayerGetShape()"); 
  return MS_FAILURE;
 }

int msInformixSpatialLayerGetExtent(layerObj *layer, rectObj *extent)
 {
  msSetError(MS_INFORMIXSPATIALERR, "InformixSpatial is not supported", "msInformixSpatialLayerGetExtent()"); 
  return MS_FAILURE;
 }

int msInformixSpatialLayerInitItemInfo(layerObj *layer)
 {
  msSetError(MS_INFORMIXSPATIALERR, "InformixSpatial is not supported", "msInformixSpatialLayerInitItemInfo()"); 
  return MS_FAILURE;
 }

void msInformixSpatialLayerFreeItemInfo(layerObj *layer)
 {
  msSetError(MS_INFORMIXSPATIALERR, "InformixSpatial is not supported", "msInformixSpatialLayerFreeItemInfo()"); 
 }

int msInformixSpatialLayerGetAutoStyle(mapObj *map, layerObj *layer, classObj *c, int tile, long record)
 {
  msSetError(MS_INFORMIXSPATIALERR, "InformixSpatial is not supported", "msInformixSpatialLayerGetAutoStyle()"); 
  return MS_FAILURE;
 }

#endif

int msInformixSpatialLayerGetShapeVT(layerObj *layer, shapeObj *shape, int tile, long record)
 {
  fprintf(stderr, "msInformixSpatialLayerGetShapeVT\n");
  return msInformixSpatialLayerGetShape(layer, shape, record);
 }

int msInformixSpatialLayerInitializeVirtualTable(layerObj *layer)
 {
  msDebug("msInformixSpatialLayerInitializeVirtualTable\n");

  // assert(layer != NULL);
  // assert(layer->vtable != NULL);

  layer->vtable->LayerInitItemInfo = msInformixSpatialLayerInitItemInfo;
  layer->vtable->LayerFreeItemInfo = msInformixSpatialLayerFreeItemInfo;
  layer->vtable->LayerOpen = msInformixSpatialLayerOpen;
  layer->vtable->LayerIsOpen = msInformixSpatialLayerIsOpen;
  layer->vtable->LayerWhichShapes = msInformixSpatialLayerWhichShapes;
  layer->vtable->LayerNextShape = msInformixSpatialLayerNextShape;
  layer->vtable->LayerGetShape = msInformixSpatialLayerGetShapeVT;

  layer->vtable->LayerClose = msInformixSpatialLayerClose;
  layer->vtable->LayerGetItems = msInformixSpatialLayerGetItems;
  layer->vtable->LayerGetExtent = msInformixSpatialLayerGetExtent;

  /* layer->vtable->LayerGetAutoStyle, use default */

  layer->vtable->LayerCloseConnection = msInformixSpatialLayerClose;

  layer->vtable->LayerApplyFilterToLayer = msLayerApplyCondSQLFilterToLayer;

  layer->vtable->LayerSetTimeFilter = msLayerMakePlainTimeFilter;
  /* layer->vtable->LayerCreateItems, use default */
  /* layer->vtable->LayerGetNumFeatures, use default */

  fprintf(stderr, "4\n");
  return MS_SUCCESS;
 }
